
#include "app.h"

static const char *g_scenePath;

G3D_START_AT_MAIN();

int main(int argc, const char *argv[])
{
    GApp::Settings s;
    s.window.caption = "CS224 Project 3: Photon";

    if (argc > 1)
    {
        g_scenePath = argv[1];
    }
    else
    {
        g_scenePath = "/course/cs224/lib/g3d9/G3D/data/scene/CornellBox-Sphere.scn.any";
        printf("No scene specified, using default: %s\n", g_scenePath);
    }

    return App(s).run();
}


App::App(const GApp::Settings &settings)
    : GApp(settings),
      stage(App::IDLE),
      view(App::DEFAULT)
{ }

App::~App() { }

/** Bumps a position */
static Vector3 bump(Vector3 pos, Vector3 dir, Vector3 normal)
{
    return pos + sign(dir.dot(normal)) * 1e-4 * normal;
}

/** Bumps a ray in place */
static void bump(Ray &ray, shared_ptr<Surfel> surf)
{
    ray.set(bump(ray.origin(), ray.direction(), surf->shadingNormal), 
            ray.direction());
}

/** Computes ray.origin + t * ray.direction, bumps the point according to the
  * normal vector, and returns it.
  */
static Vector3 bump(Ray &ray, float t, Vector3 normal)
{
    return bump(ray.origin() + t * ray.direction(), ray.direction(), normal);
}

void App::scatter()
{
    // TODO emit a photon and let it bounce around
    //      The function World::emit() will help you emit a photon
    //      Surfel::scatter() will help you bounce it
}

Radiance3 App::direct(shared_ptr<Surfel> surf, Vector3 wo)
{
    Radiance3 rad;

    shared_ptr<Surfel> light;
    float P_light;
    float area;

    for (int i = 0; i < DIRECT_SAMPLES; ++i)
    {
        m_world.emissivePoint(m_random, light, P_light, area);

        Vector3 wi = light->position - surf->position;;
        float dist = wi.length();
        if (dist < 1e-4)
            continue;
        wi /= dist;

        if (m_world.lineOfSight(surf->position + wi * 1e-4, light->position))
        {
            rad += light->emittedRadiance(-wi) / (pif() * area)
                 * surf->finiteScatteringDensity(wi, wo)
                 * max(0.f, wi.dot(surf->shadingNormal))
                 * max(0.f, -wi.dot(light->shadingNormal))
                 / (dist * dist)
                 / P_light
                 / DIRECT_SAMPLES;
        }
    }

    return rad;
}

Radiance3 App::impulse(shared_ptr<Surfel> surf, Vector3 wo, int depth)
{
    if (!--depth) 
        return Radiance3::zero();

    Surfel::ImpulseArray imp;
    surf->getImpulses(PathDirection::EYE_TO_SOURCE, wo, imp);

    Radiance3 rad;
    for (int i = 0; i < imp.size(); ++i)
    {
        Ray ray(surf->position, imp[i].direction);
        bump(ray, surf);

        rad += imp[i].magnitude * trace(ray, depth);
    }

    return rad;
}

/** The photon map kernel (a cone filter).
  * @param dist The distance between the point being sampled and a photon
  * @return     The kernel weight for this photon
  */
static float cone(float dist)
{
    static const float volume = pi() * square(GATHER_RADIUS) / 3;
    static const float normalize = 1.f / volume;

    float height = 1.f - dist / GATHER_RADIUS;
    return height * normalize;
}

Radiance3 App::diffuse(shared_ptr<Surfel> surf, Vector3 wo)
{
    // In line with the path demo, ignore diffuse interrfelection for specular
    // surfaces. 
    Surfel::ImpulseArray imp;
    surf->getImpulses(PathDirection::EYE_TO_SOURCE, wo, imp);
    if (imp.size() > 0)
        return Radiance3::zero();

    Radiance3 rad;  
    
    // TODO compute indirect illumination using photon map
    // Hint: PhotonMap is an alias for the G3D type PointHashGrid.
    //       SphereIterator lets you iterate through all photons inside a sphere
    // Hint: Your gather radius should be GATHER_RADIUS, defined in app.h
    // Hint: Use the cone() function above as your photon map's kernel

    return rad;
}

Radiance3 App::trace(const Ray &ray, int depth)
{
    Radiance3 final;

    float dist = 0;
    shared_ptr<Surfel> surf;
    m_world.intersect(ray, dist, surf);

    if (surf)
    {
        Point3 loc = ray.origin() + ray.direction() * dist;
        Point3 eye = ray.origin();
        Vector3 wo = -ray.direction();

        final += surf->emittedRadiance(wo)  
               + direct(surf, wo)
               + diffuse(surf, wo)
               + impulse(surf, wo, depth);
    }

    return final;
}

void App::buildPhotonMap()
{
    for (int i = 0; i < NUM_PHOTONS; ++i)
    {
        printf("\rBuilding photon map ... %.2f%%", 100.f * i / NUM_PHOTONS);
        scatter();
    }
    printf("\rBuilding photon map ... done       \n");
}

void App::traceCallback(int x, int y)
{
    Ray ray = m_world.camera()->worldRay(x + .5f, y + .5f, m_canvas->rect2DBounds());
    m_canvas->set(x, y, trace(ray, MAX_DEPTH));
}

static void dispatcher(void *arg)
{
    App *self = (App*)arg;

    int w = self->window()->width(),
        h = self->window()->height();

    self->stage = App::SCATTERING;
    self->buildPhotonMap();

    printf("Rendering ...");
    fflush(stdout);
    self->stage = App::GATHERING;
    GThread::runConcurrently2D(Point2int32(0, 0),
                               Point2int32(w, h),
                               self,
                               &App::traceCallback);
    printf("done\n");

    self->stage = App::IDLE;
}

void App::onInit()
{
    GApp::showRenderingStats = false;

    m_world.load(g_scenePath);

    int w = window()->width(), 
        h = window()->height();

    m_canvas = Image3::createEmpty(w, h);
    m_gamma = Texture::createEmpty("m_gamma", w, h, m_canvas->format());

    m_dispatch = GThread::create("dispatcher", dispatcher, this);
    m_dispatch->start();
}

void App::onCleanup()
{
    m_world.unload();
}

void App::onGraphics(RenderDevice *dev, 
                     Array<shared_ptr<Surface> >& posed3D,
                     Array<shared_ptr<Surface2D> >& posed2D)
{
    View v = this->view;
    if (v == DEFAULT)
        v = stage == SCATTERING ? PHOTONMAP : RENDITION;

    if (v == PHOTONMAP) 
    {
        dev->setColorClearValue(Color4(0.0, 0.0, 0.0, 0.0));
        dev->clear();
        m_photons.render(dev, &m_world);
    }
    else
    {
        m_gamma->update(m_canvas);

        FilmSettings s;
        s.setAntialiasingEnabled(false);
        s.setBloomStrength(0);
        s.setGamma(2.060);
        s.setVignetteTopStrength(0);
        s.setVignetteBottomStrength(0);
        m_film->exposeAndRender(renderDevice, s, m_gamma);
    }
}

void App::onUserInput(UserInput* input)
{
    if (input->keyReleased(GKey('1')))
        this->view = PHOTONMAP;

    if (input->keyReleased(GKey('2')))
        this->view = RENDITION;

    if (input->keyReleased(GKey('0')))
        this->view = DEFAULT;
}

