
#include "app.h"
#include "world.h"

World::World() { }

World::~World() { }

void World::load(const std::string &path)
{
    printf("Loading scene %s...\n", path.c_str());

    Any scene;
    scene.load(path);

    // Read the model table
    debugAssert(scene.containsKey("models"));
    const Table<std::string, Any> models = scene["models"].table();

    // Dump it to stdout
    printf("%d model(s)\n", (int)models.size());
    for (int i = 0; i < (int)models.size(); ++i)
        printf("    %s\n", models.getKeys()[i].c_str());

    // Read the entity table
    debugAssert(scene.containsKey("entities"));
    const Table<std::string, Any> &entities = scene["entities"].table();

    // Parse entities
    printf("%d entities(s)\n", (int)entities.size());
    Array<Surface::Ref> geometry;
    for (int i = 0; i < (int)entities.size(); ++i)
    {
        std::string key = entities.getKeys()[i];
        Any e = entities[key];
        std::string type = e.name();

        printf("    %s (%s) ... ", key.c_str(), type.c_str());

        if (type == "Camera") 
        {
            AnyTableReader props(e);
            m_camera = Camera::create(type, props);

            printf("done\n");
        }
        else if (type == "Light")
        {
            printf("ignored (only emitters are used as lights in photon)\n");
        }
        else if (type == "VisibleEntity")
        {
            const Table<std::string, Any> &props = e.table();

            // Read the model from disk
            ArticulatedModel::Ref model = 
                ArticulatedModel::create(models[props["model"]]);

            // Pose it in world space
            Vector3 pos = Vector3::zero();
            if (props.containsKey("position"))
                pos = Vector3(props["position"]);

            Array<Surface::Ref> posed;
            model->pose(posed, CFrame(pos));

            // Add it to the scene
            for (int i = 0; i < posed.size(); ++i)
                geometry.append(posed[i]);

            printf("done\n");
        }
        else
        {
            printf("ignored (unknown entity type)\n");
        }
    }    

    // Build bounding interval hierarchy for scene geometry
    Array<Tri> triArray;
    
    Surface::getTris(geometry, m_verts, triArray);
    for (int i = 0; i < triArray.size(); ++i)
    {
        triArray[i].material()->setStorage(MOVE_TO_CPU);

        // Check if this triangle emits light
        Material::Ref m = triArray[i].material()->resolve();
        if (m)
        {
            UniversalMaterial::Ref mtl = 
                dynamic_pointer_cast<UniversalMaterial>(m);

            if (mtl->emissive().notBlack())
                m_emit.append(triArray[i]);
        }
    }

    TriTree::Settings s;
    s.algorithm = TriTree::MEAN_EXTENT;
    m_tris.setContents(triArray, m_verts, s);

    printf("%d light-emitting triangle(s) in scene.\n", (int)m_emit.size());
}

void World::unload()
{
    m_tris.clear();
}

Camera::Ref World::camera()
{
    return m_camera;
}

void World::emissivePoint(Random &random, shared_ptr<Surfel> &surf, float &prob, float &area)
{
    // Pick an emissive triangle uniformly at random
    int i = random.integer(0, m_emit.size() - 1);
    const Tri& tri = m_emit[i];

    // Pick a point in that triangle uniformly at random
    // http://books.google.com/books?id=fvA7zLEFWZgC&pg=PA24#v=onepage&q&f=false
    float s = random.uniform(), 
          t = random.uniform(), 
          sqrtT = sqrt(t),
          a = (1.f - sqrtT),
          b = (1.f - s) * sqrtT,
          c = s * sqrtT;

    Vector3 point = tri.position(m_verts, 0) * a
                  + tri.position(m_verts, 1) * b
                  + tri.position(m_verts, 2) * c;

    Vector3 normal = tri.normal(m_verts, 0) * a
                   + tri.normal(m_verts, 1) * b
                   + tri.normal(m_verts, 2) * c;
    normal = normal.direction();

    // XXX semi-hacky, slow way to get a surfel for this point
    //     there has to be a better way ...
    Ray ray(point + normal * 1e-4, -normal);
    float dist = 0;
    intersect(ray, dist, surf);

    prob = 1.f / m_emit.size() / tri.area();

    area = tri.area();
}

bool World::emit(Random &random, Photon &photon, shared_ptr<Surfel> &surf)
{
    // Select the point of emission
    shared_ptr<Surfel> light;
    float prob;
    float area;
    emissivePoint(random, light, prob, area);

    // Shoot the photon somewhere into the scene
    Vector3 dir;
    float dist;

    dir = Vector3::cosHemiRandom(light->shadingNormal, random);
    intersect(Ray(light->position + light->geometricNormal * 1e-4, dir), dist, surf);

    if (!surf)
        return false;
    
    // Store the photon
    photon.position = surf->position;
    photon.wi = -dir;
    photon.power = light->emittedRadiance(dir)
                 / NUM_PHOTONS 
                 * m_emit.size();

    return true;
}

void World::intersect(const Ray &ray, float &dist, shared_ptr<Surfel> &surf)
{
    dist = INFINITY; // No max distance
    surf = m_tris.intersectRay(ray, dist);
}

bool World::lineOfSight(const Vector3 &beg, const Vector3 &end)
{
    Vector3 d = end - beg;
    float dist = d.length();
    if (dist < 1e-4)
        return false;

    Ray ray = Ray::fromOriginAndDirection(beg, d / dist);

    // exitOnAnyHit -- Find any intersection, not the first (early exit)
    // twoSidedTest -- Don't cull back-facing triangles from intersectino test
    static const bool exitOnAnyHit = true, twoSidedTest = true;

    Tri::Intersector intersector;
    return !m_tris.intersectRay(ray, intersector, dist, exitOnAnyHit, twoSidedTest);
}

